(***************************************************************)
(*                                                             *)
(*                 camlgrenouille for Linux                    *)
(*                                                             *)
(*             Copyright 2001-2003 Alan Schmitt                *)
(*              alan.schmitt@polytechnique.org                 *)
(*                                                             *)
(* This file is distributed under the terms of the GNU General *)
(* Public License.                                             *)
(*                                                             *)
(***************************************************************)

open Unix
open Debug

type fill_buffer_type = string -> int -> int
type read_buffer_type = string -> unit

type time_or_block =
  | Time of float
  | Block

let buf_size = 32 * 1024

let fill_buffer_from_string str =
  let len = String.length str in
  let start = ref 0 in
  let left = ref len in
  let f abuf abuf_size =
    if !left > abuf_size then (
      String.blit str !start abuf 0 abuf_size;
      start := !start + abuf_size;
      left := !left - abuf_size;
      abuf_size
     )
    else (
      String.blit str !start abuf 0 !left;
      let old = !left in
      left := 0;
      old
     )
  in
  f 
  
let fill_buffer_from_file_descr d abuf abuf_size =
  let n = input d abuf 0 abuf_size in
  n

let buf_write time conn fill_buffer =
  let buf = String.create buf_size in
  let rec write_rec pos nb_to_wr =
    let nb = 
      if nb_to_wr = 0 then (
	(* dprint Debug "filling buffer"; *)
	fill_buffer buf buf_size
       )
      else
	nb_to_wr
    in
    if nb <> 0 then (
      (* dprint Debug ("attempt to send: "^(string_of_int nb)^" starting at "^(string_of_int pos)); *)
      let nb_written = Timeout.timeout time (write conn buf pos) nb in
      (* dprint Debug ("sent: "^(string_of_int nb_written)); *)
      if nb_written <> nb then 
	write_rec (pos+nb_written) (nb - nb_written)
      else
	write_rec 0 0
     )
  in
  write_rec 0 0

let rec buf_read tb conn func_to_apply =
  let buf = String.create buf_size in
  let rec read_rec () =
    (* dprint Debug "in Bufrw.read_rec"; *)
    let nb = 
      match tb with
      | Time time -> Timeout.timeout time (read conn buf 0) buf_size 
      | Block ->
          (try
            read conn buf 0 buf_size
          with
          | Unix_error ((EWOULDBLOCK | EAGAIN), _, _) -> 0)
    in
    (* dprint Debug ("read: "^(string_of_int nb)); *)
    match nb with
    | 0 -> ()
    | n when n = buf_size -> 
(*	print_endline "read a full buffer ;-)"; *)
	func_to_apply buf; 
	read_rec ()
    | n -> 
	func_to_apply (String.sub buf 0 nb); 
	read_rec ()
  in
  read_rec ()

let buf_read_fast time conn func_to_apply buf pos =
  let rec read_rec pos =
    let nb = Timeout.timeout time (read conn buf 0) buf_size in
    match nb with
    | 0 -> ()
    | n when n = buf_size -> 
(*        print_endline "read a full buffer ;-)";*)
        let pos = func_to_apply pos n in
        read_rec pos
    | n -> 
        let pos = func_to_apply pos nb in
        read_rec pos
  in
  read_rec pos
  
