(************************************************************************)
(*                                                                      *)
(*                        Caml Development Kit                          *)
(*                                                                      *)
(*                 Projet Moscova, INRIA Rocquencourt                   *)
(*                                                                      *)
(************************************************************************)

(*c functions for timed out buffered read from and write to Unix
   file descriptors *)

type fill_buffer_type = string -> int -> int
(*d the type of functions to fill buffers. [my_fill_buffer buf size] 
   takes a buffer [buf] of size [size] and returns the number of
   bytes copied in the buffer. Such a function should return 0 when there is
   no more data available.
*)

type read_buffer_type = string -> unit
(*d the type of functions to read buffers *)

type time_or_block =
  | Time of float
  | Block
(*d to specify how to behave on input blocking:
    [Time t] raises a [Timeout.Timeout] exception after blocking for t seconds
    [Block] either blocks, or return immediately, depending on the configuration
    of the file descriptor. *)

val fill_buffer_from_string : string -> fill_buffer_type
(*d to fill a buffer from a string *)

val fill_buffer_from_file_descr : in_channel -> fill_buffer_type
(*d to fill a buffer from the contents of a file *)

val buf_write : float -> Unix.file_descr -> fill_buffer_type -> unit
(*d [buf_write timeout fd my_fill_buffer] writes the data provided by
   [my_fill_buffer] on the file descriptor [fd]. [raise Timeout.Timeout]
   if writing one buffer takes longer than [timeout] seconds.
*)

val buf_read : time_or_block -> Unix.file_descr -> read_buffer_type -> unit
(*d [buf_read tb fd my_read_buffer] reads data from a file
  descriptor [fd] until no more data is available, applying
  [my_read_buffer] to each buffered read. May [raise Timeout.Timeout] if
  [tb] is [Time t].
*)

  
val buf_size : int
(*d [buf_size] is the minimum size of a buffer for [buf_read_fast]. *)

val buf_read_fast: float ->
  Unix.file_descr -> (int -> int -> int) -> string -> int -> unit
(*d [buf_read_fast timeout fd my_read_buffer buf pos] reads date from
  a file descriptor [fd] until no more data is available, applying
  [my_read_buffer] to each buffered read. [raise Timeout.Timeout] if
  reading one buffer takes longer than [timeout] seconds. It fills [buf]
  starting at [pos]. [my_read_buffer] receives the position and length of
  the date stored in [buf] and should return the new position for filling.
*)
