(***************************************************************)
(*                                                             *)
(*                      camlgrenouille                         *)
(*                                                             *)
(*             Copyright 2001-2005 Alan Schmitt                *)
(*              alan.schmitt@polytechnique.org                 *)
(*                                                             *)
(* This file is distributed under the terms of the GNU General *)
(* Public License.                                             *)
(*                                                             *)
(***************************************************************)

open Debug
open Config
open Sysconf
open Event_list
open Abort

exception Grenouille_command of string * string

let parse_value s =
  dprint Debug s;
  let ind = String.index s '=' in
  String.sub s 0 ind, String.sub s (ind+1) ((String.length s) - ind - 1)

let end_with_slash s =
  let l = String.length s in
    (l > 0) && (String.get s (l-1) = '/')

exception Grenouille_exc of string

let min_time = ref 0
let connection_failures = ref 0

let compute_delay () =
  if !min_time = 0 then
    ((if !connection_failures > 15 then 15 else !connection_failures) * 5 + 2) * 60
  else
    (!min_time / 5) * (if !connection_failures > 5 then 5 else !connection_failures)
                              
let id = ref ""
let alt_id = ref ""

let use_alt_server = ref false

type result_store = {
    mutable id : string;
    
    mutable dl_host : string;
    mutable dl_path : string;
    mutable dl_file : string;
    mutable dl_frequency : string;
    mutable dl_max : string;
    mutable dl_ratiomaxdl : string;
    mutable dl_ratiomaxul : string;

    mutable ul_host : string;
    mutable ul_path : string;
    mutable ul_file : string;
    mutable ul_size : string;
    mutable ul_frequency : string;
    mutable ul_max : string;
    mutable ul_ratiomaxdl : string;
    mutable ul_ratiomaxul : string;
    
    mutable ping_host : string;
    mutable ping_quantity : string;
    mutable ping_frequency : string;
    mutable ping_min : string;
    mutable ping_max : string;
    mutable ping_string : string;
    
    mutable breakdown_host : string;
    mutable breakdown_frequency : string;

    mutable time_daylightsaving : string;
    mutable time_zone : string;

    mutable config_timetolive : string;

    mutable rclient : string;
    mutable rversion : string;
    mutable rbeta : string;
    mutable rbuild : string;
    mutable rsystem : string;
    
    mutable ok : bool;
    mutable mark : int option;
    mutable warning : string;
  } 

let new_empty_rec_store () = {
  id = "";

  dl_host = "";
  dl_path = "";
  dl_file = "";
  dl_frequency = "";
  dl_max = "";
  dl_ratiomaxdl = "";
  dl_ratiomaxul = "";

  ul_host = "";
  ul_path = "";
  ul_file = "";
  ul_size = "";
  ul_frequency = "";
  ul_max = "";
  ul_ratiomaxdl = "";
  ul_ratiomaxul = "";
  
  ping_host = "";
  ping_quantity = "";
  ping_frequency = "";
  ping_min = "";
  ping_max = "";
  ping_string = "";
  
  breakdown_host = "";
  breakdown_frequency = "";

  time_daylightsaving = "";
  time_zone = "";

  config_timetolive = "";

  rclient = "";
  rversion = "";
  rbeta = "";
  rbuild = "";
  rsystem = "";

  ok = false;
  mark = None;
  warning = "";
}
      
let parse_value_list l = 
  let rec_store = new_empty_rec_store () in
  let rec pvrec = 
    function
      | [] -> rec_store
      | v :: l -> (
	  match v with
	  | "error",s -> raise (Grenouille_exc s)
	  | "warning", s -> rec_store.warning <- s
	  | "id",s -> rec_store.id <- s
	  | "dl_host",s -> rec_store.dl_host <- s
	  | "dl_path",s -> rec_store.dl_path <- s
	  | "dl_file",s -> rec_store.dl_file <- s
	  | "dl_frequency",s -> rec_store.dl_frequency <- s
	  | "dl_max",s -> rec_store.dl_max <- s
	  | "dl_ratiomaxdl",s -> rec_store.dl_ratiomaxdl <- s
	  | "dl_ratiomaxul",s -> rec_store.dl_ratiomaxul <- s
	  | "ul_host",s -> rec_store.ul_host <- s
	  | "ul_path",s -> rec_store.ul_path <- s
	  | "ul_file",s -> rec_store.ul_file <- s
	  | "ul_size",s -> rec_store.ul_size <- s
	  | "ul_frequency",s -> rec_store.ul_frequency <- s
	  | "ul_max",s -> rec_store.ul_max <- s
	  | "ul_ratiomaxdl",s -> rec_store.ul_ratiomaxdl <- s
	  | "ul_ratiomaxul",s -> rec_store.ul_ratiomaxul <- s
	  | "ping_host",s -> rec_store.ping_host <- s
	  | "ping_quantity",s -> rec_store.ping_quantity <- s
	  | "ping_frequency",s -> rec_store.ping_frequency <- s
	  | "ping_min",s -> rec_store.ping_min <- s
	  | "ping_max",s -> rec_store.ping_max <- s
	  | "ping_string",s -> rec_store.ping_string <- s
	  | "breakdown_host",s -> rec_store.breakdown_host <- s
	  | "breakdown_frequency",s -> rec_store.breakdown_frequency <- s
	  | "time_daylightsaving",s -> rec_store.time_daylightsaving <- s
	  | "time_zone",s -> rec_store.time_zone <- s
	  | "config_timetolive",s -> rec_store.config_timetolive <- s
	  | "client",s -> rec_store.rclient <- s
	  | "version",s -> rec_store.rversion <- s
	  | "beta",s -> rec_store.rbeta <- s
	  | "build",s -> rec_store.rbuild <- s
	  | "system",s -> rec_store.rsystem <- s
          | "mark", s -> rec_store.mark <- Some (int_of_string s)
	  | "result","OK" -> rec_store.ok <- true
	  | s1,s2 -> dprint Normal ("Argument "^s1^"="^s2^" inconnu")
	 );
	  pvrec l
  in
  pvrec l

let process_command cmd_str alt =
  let start_cmd =
    "GET " ^ 
    (if alt then usrconf.alt_grenouille_file else usrconf.grenouille_file)
    ^ "?command=" in
	let end_cmd = " HTTP/1.0\r\n" ^
		"User-Agent: grenouille camlgrenouille/" ^ sysconf.version_maj ^ "."
		^ sysconf.version_min ^ " (" ^ sysconf.system ^ ") \r\n" ^
		"Host: " ^ 
		(if alt then usrconf.alt_grenouille_server else usrconf.grenouille_server)
		^ "\r\n" ^ "\r\n"
	in
  let cmd = start_cmd ^ cmd_str ^ end_cmd in
  dprint Debug "\n";
	dprint Debug ("Processing command: "^cmd);
  let headers,content =
    Http.request cmd
      (if alt then usrconf.alt_grenouille_server else usrconf.grenouille_server)
      (if alt then usrconf.alt_grenouille_port else usrconf.grenouille_port)
  in
  dprint Debug "Headers:";
  ignore (List.map (fun s -> dprint Debug s) headers);
  dprint Debug "";
  dprint Debug "Content:";
  let value_list = List.map (fun s -> parse_value s) content in
  parse_value_list value_list

let say_hello pass alt =
  let opt_ip_str =
    if usrconf.detect_router then
      let plugin_str = Utils.get_plugin_str () in
      let ip_str = Utils.get_result_number (Utils.get_plugin_result plugin_str) 2 in
        dprint Debug ("L'IP trouvée par la machine est " ^ ip_str);
        ("&ip=" ^ ip_str)
    else
      ""
  in
  let hello = ("hello&username=" ^ usrconf.username ^
               "&password=" ^ pass ^
               "&client=" ^ sysconf.client ^
               "&version=" ^ sysconf.version_maj ^
               "." ^ sysconf.version_min ^
               "&system=" ^ sysconf.system ^
               (match sysconf.beta with None -> "" | Some s -> ("&beta="^s)) ^ opt_ip_str) in
  let res = process_command hello alt in
  if res.ok && (res.id <> "") 
  then begin
    if alt then
      alt_id := res.id
    else
      id := res.id
  end
  else failwith "Could not say hello"

type commands = {
    mutable get_config : bool -> unit;
    mutable download : unit -> unit;
    mutable upload : unit -> unit;
    mutable ping : unit -> unit;
    mutable breakdown : unit -> unit;
    mutable post_ping : int -> int -> float -> float -> bool -> result_store;
    mutable post_ul : float -> float -> bool -> result_store;
    mutable post_dl : float -> float -> bool -> result_store;
    mutable post_breakdown : float -> bool -> result_store;
    mutable test_download : float -> float -> float -> bool;
    mutable test_upload : float -> float -> float -> bool;
  }

let commands =
  let empty_rec = new_empty_rec_store () in 
    {
      get_config = (fun _ -> ());
      download = (fun () -> ());
      upload = (fun () -> ());
      ping = (fun () -> ());
      breakdown = (fun () -> ());
      post_ping = (fun _ _ _ _ _ -> empty_rec);
      post_ul = (fun _ _ _ -> empty_rec);
      post_dl = (fun _ _ _ -> empty_rec);
      post_breakdown = (fun _ _ -> empty_rec);
      test_download = (fun _ _ _ -> false);
      test_upload = (fun _ _ _ -> false);
    } 

let generate_test_download max_dl max_ul =
  fun base diff_in diff_out ->
    if usrconf.noratiodl then true else
    if base = 0.0 || diff_in < base then ( 
	dprint Normal ("Not enough bandwidth was used !");
	false; )
    else (
    let dl = int_of_float(((diff_in -. base) *. 100.0) /. base) in
    let ul = int_of_float(diff_out *. 100.0 /. base) in
    dprint Normal ("Download ratio: dl: "^(string_of_int dl)^"%; ul: "^(string_of_int ul)^"%");
    dl <= (max max_dl usrconf.ratiodl) && ul <= max_ul )

let generate_test_upload max_dl max_ul =
  fun base diff_in diff_out ->
    if usrconf.noratioul then true else    
    if base = 0.0 || diff_out < base then ( 
	dprint Normal ("Not enough bandwidth was used !");
	false; )
    else
    let ul = int_of_float(((diff_out -. base) *. 100.0) /. base) in
    let dl = int_of_float(diff_in *. 100.0 /. base) in
    dprint Normal ("Upload ratio: dl: "^(string_of_int dl)^"%; ul: "^(string_of_int ul)^"%");
    ul <= (max max_ul usrconf.ratioul) && dl <= max_dl

type result =
  | Ping of int * int * float
  | Download of float
  | Upload of float
  | Breakdown

type result_list = (result * float) list

let results_to_send : result_list ref = ref []
    
let add_result r =
  results_to_send := (r, Unix.time ()) :: !results_to_send

let send_result (r,t) use_alt =
  let elapsed = (Unix.time ()) -. t in
  let res, kind =
  match r with
  | Ping (sent, received, total_time) ->
      commands.post_ping sent received total_time elapsed use_alt, "ping"
  | Upload bandwidth ->
      commands.post_ul bandwidth elapsed use_alt, "upload"
  | Download bandwidth ->
      commands.post_dl bandwidth elapsed use_alt, "download"
  | Breakdown ->
      commands.post_breakdown elapsed use_alt, "breakdown"
  in
  if not res.ok then failwith "OK not received: problem posting";
  (match res.mark with
  | None -> ()
  | Some i -> dprint Normal ("Quality of "^kind^" (between 1 and 5): "^(string_of_int i)));
  if res.warning <> "" then (
    dprint Normal ("received warning, result thrown away: "^res.warning)
   )

let do_post_results () =
  let rec iter_send = function
    | [] -> ()
    | res :: l ->
        send_result res false;
        if !use_alt_server then send_result res true;
        results_to_send := l;
        iter_send l
  in
    try
      if usrconf.post_results then (
        iter_send !results_to_send;
        connection_failures := 0;
        dprint Normal "Results sent" )
      else
        dprint Normal "config files says not to send results";
      Timed_events.add_event_unique PostResults (Future.set_time usrconf.normal_result_delay)
    with
      | Aborting -> raise Aborting
      | e ->
          dprint Normal ("Problem sending results:\n" ^ Printexc.to_string e);
          incr connection_failures;
          let delay = compute_delay () in
          let date = Future.set_time delay in
            Timed_events.add_event_unique PostResults date;
            Timed_events.delay_connections false date delay

let do_dummy () =
  Printf.printf "DUMMY\n";
  flush stdout;
  Timed_events.add_event_unique DoDummy (Future.set_time 1000)
          
let get_last_version id =
  let res = process_command ("get_last_version&id=" ^ id) false in
  if not res.ok then raise (Grenouille_command ("get_last_version","ok false"));
  let verf = float_of_string res.rversion in
  let cur_verf = float_of_string (sysconf.version_maj ^ "." ^ sysconf.version_min) in
  dprint Normal "";
  if verf <> cur_verf then
    if verf > cur_verf then
      dprint Normal ("New version available: " ^ (Printf.sprintf "%.2f" verf) ^
		     (if res.rbeta <> "255" then "b"^res.rbeta else ""))
    else dprint Normal "Waow, you're using a version more recent than the one announced on grenouille.com ;-)"
  else
    dprint Normal "Congratulations, you are using the latest version of camlgrenouille";
  dprint Normal ""
      
let download dl_host dl_path dl_file dl_ratiomaxdl dl_ratiomaxul freq () =
  try
    let args = 
      if (usrconf.ftp_password <> "") then [Ftp.User "anonymous"; Ftp.Password usrconf.ftp_password]
      else []
    in
    let conn = Ftp.connect dl_host args in
    let dest = ("/tmp/" ^ dl_file) in
    let sep = if end_with_slash dl_path then "" else "/" in
    (try
      let exp_size = Ftp.size conn (dl_path^sep^dl_file) in
      dprint Debug ("expected size: "^(string_of_int exp_size));
      let out = open_out_bin dest in
      let before_in, before_out = Utils.get_net_stats () in
      let time =
	try
	  let t = 
	    Ftp.get conn (dl_path^sep^dl_file) (output_string out) (if usrconf.passive then [Ftp.Passive] else [])
	  in
	  myign close_out out;
	  t
	with
	| e -> myign close_out out; raise e
      in
      let after_in, after_out = Utils.get_net_stats () in
      let real_size = (Unix.stat dest).Unix.st_size in
      if real_size <> exp_size then (
	dprint Debug ("we got: "^(string_of_int real_size)^"; we wanted: "^(string_of_int exp_size));
	failwith "downloaded and expected size are different");
      dprint Debug ("download time: "^(string_of_float time));
      dprint Debug ("download size: "^(string_of_int exp_size));
      let bandwidth = (float_of_int exp_size)/. 1024. /. time in
      dprint Debug ("bandwidth: "^(string_of_float bandwidth));
      dprint Debug ("download, in: "^(string_of_float before_in)^" -> "^(string_of_float after_in)^"; out: "^(string_of_float before_out)^" -> "^(string_of_float after_out));
      if (commands.test_download (float_of_int exp_size) (after_in -. before_in) (after_out -. before_out))
      then (
	dprint Normal ("Download: "^(string_of_float bandwidth)^" ko/s");
	add_result (Download bandwidth);
        Timed_events.done_successful_dl ();
	myign (exec_out usrconf.dl_out) (string_of_float bandwidth)
       )
      else
	dprint Normal "Download result thrown away, the network was used too much OR WRONG INTERFACE SETTING";
      Timed_events.add_event_unique DoDownload (Future.set_time freq)
    with
    | Aborting -> 
	Ftp.quit conn;
	raise Aborting
    | e ->
	dprint Normal ("Problem testing download:\n" ^ Printexc.to_string e);
	Ftp.quit conn;
	Timed_events.add_event_unique DoDownload (Future.set_time (freq / 5)));
    myign Ftp.quit conn;
    myign Unix.unlink ("/tmp/" ^ dl_file)
  with
  | Aborting -> raise Aborting
  | e -> 
      dprint Normal ("Could not connect to ftp:\n" ^ Printexc.to_string e);
      Timed_events.add_event_unique DoDownload (Future.set_time (freq / 5))
	
let upload ul_max ul_host ul_path ul_file ul_size ul_ratiomaxdl ul_ratiomaxul freq () =
  try
    let conn = Ftp.connect ul_host [] in
    let tmp_file = "/tmp/"^ul_file in
    (try
      Utils.generate_random_file tmp_file (ul_size * 1024);
      Ftp.cd conn ul_path;
      let before_in, before_out = Utils.get_net_stats () in
      let descr = open_in_bin tmp_file in
      let time =
	try
	  let t = Ftp.put conn (Bufrw.fill_buffer_from_file_descr descr) ul_file (if usrconf.passive then [Ftp.PassivePut] else []) in
	  myign close_in descr; t
	with
	| e -> myign close_in descr; raise e
      in
      let after_in, after_out = Utils.get_net_stats () in
      dprint Debug ("upload, in: "^(string_of_float before_in)^" -> "^(string_of_float after_in)^"; out: "^(string_of_float before_out)^" -> "^(string_of_float after_out));
      dprint Debug ("uploaded: "^(string_of_int ul_size)^" ko in "^(string_of_float time)^" seconds");
      if (commands.test_upload (float_of_int ((ul_size * 1024))) (after_in -. before_in) (after_out -. before_out)) then (
	let bandwidth =  (float_of_int ul_size) /. time in
	dprint Normal ("Upload: "^(string_of_float bandwidth)^" ko/s");
	if bandwidth > ul_max then 
	  dprint Debug "SUPER UPLOAD !!!";
	add_result (Upload bandwidth);
	myign (exec_out usrconf.ul_out) (string_of_float bandwidth)
       )
      else
	dprint Normal "Upload result thrown away, the network was used too much OR WRONG INTERFACE SETTING";
      Timed_events.add_event_unique DoUpload (Future.set_time freq);
    with
    | Aborting -> 
	Ftp.quit conn;
        myign Unix.unlink tmp_file;
	raise Aborting
    | e ->
	dprint Normal ("Problem testing upload:\n" ^ Printexc.to_string e);
	Ftp.quit conn;
	Timed_events.add_event_unique DoUpload (Future.set_time (freq / 5)));
    (try
      Ftp.rm conn ul_file
    with
    | Aborting -> 
	Ftp.quit conn;
        myign Unix.unlink tmp_file;
	raise Aborting
    | e ->
	dprint Debug ("problem erasing remote file:\n"^ Printexc.to_string e));
    myign Ftp.quit conn;
    myign Unix.unlink tmp_file
  with
  | Aborting -> raise Aborting
  | e -> 
      dprint Normal ("Could not connect to ftp:\n" ^ Printexc.to_string e);
      Timed_events.add_event_unique DoUpload (Future.set_time (freq / 5))

      
let ping freq freq_dl host count str () =
  try
    let t_last_dl = Timed_events.time_since_last_dl () in
    if t_last_dl > freq_dl then begin
      dprint Normal ("Ping not tested since last successful download occured "^(string_of_int t_last_dl)^" seconds ago");
      Timed_events.add_event_unique DoPing (Future.set_time (freq/5))
    end else begin
      let ping_size = string_of_int (String.length str) in
      let trans, reic, time = 
        Timeout.timeout 60.0 (fun () -> 
          dprint Debug "Testing ping with timeout";
          Utils.ping host count ping_size) ()
      in
      dprint Normal ("ping: transmitted: "^(string_of_int trans)^"; received: "^(string_of_int reic)^"; average time: "^(string_of_float (time/. (float_of_int reic)))^" ms");
      add_result (Ping (trans, reic, time));
      myign (exec_out usrconf.ping_out) ((string_of_int trans) ^ " " ^ (string_of_int reic) ^ " " ^ (string_of_float time));  
      Timed_events.add_event_unique DoPing (Future.set_time freq)
    end
  with
  | Aborting -> raise Aborting
  | e ->
      dprint Normal ("Problem testing ping (the host may be down ...):\n" ^ Printexc.to_string e);
      Timed_events.add_event_unique DoPing (Future.set_time (freq / 5))
	
exception Link_up

let remove_initial_space s =
  let cnt = ref 0 in
  while s.[!cnt] = ' ' 
  do
    incr cnt
  done;
  String.sub s (!cnt) ((String.length s) - (!cnt))

let generate_list_from_commas s =
  let rec get_next s =
    try
      let ind = String.index s ',' in
      (remove_initial_space (String.sub s 0 ind)) :: (get_next (String.sub s (ind+1) ((String.length s) - ind - 1)))
    with
    | Not_found -> [remove_initial_space s]
  in
  get_next s

let breakdown host_list freq () =
  let test_one_host h =
    let _,r,_ =
      try
        Timeout.timeout 60.0 ( fun () ->
	  Utils.ping h "10" "0") ()
      with
      | Aborting -> raise Aborting
      | e -> 
	  dprint Normal ("Breakdown: ping failed for "^h^": "^(Printexc.to_string e));
	  0,0,0.
    in
    if r > 0 then raise Link_up
  in
  let res =
    try
      List.iter test_one_host host_list;
      false
    with
    | Link_up -> true
  in
  if not res then (
    dprint Normal "Breakdown detected";
    myign (exec_out usrconf.breakdown_out) "";
    add_result Breakdown)
  else 
    dprint Normal "The network is up and running";
  Timed_events.add_event_unique DoBreakdown (Future.set_time freq)

let err_dl = ref false
let err_ul = ref false
let err_ping = ref false
let err_bk = ref false
  
let get_config id alt =
  try
    let res = process_command ("get_config&id="^id) alt in
    if not res.ok then raise (Grenouille_command ("get_config","ok false"));

    if alt then () else begin
    
  (* download config error raising *)
    (try 
      if res.dl_host = "" then raise (Grenouille_command ("get_config","dl_host not set"));
      if res.dl_path = "" then raise (Grenouille_command ("get_config","dl_path not set"));
      if res.dl_file = "" then raise (Grenouille_command ("get_config","dl_file not set"));
      if res.dl_frequency = "" then raise (Grenouille_command ("get_config","dl_frequency not set"));
      min_time := int_of_string res.dl_frequency;
      if res.dl_ratiomaxdl = "" then raise (Grenouille_command ("get_config","dl_ratiomaxdl not set"));
      if res.dl_ratiomaxul = "" then raise (Grenouille_command ("get_config","dl_ratiomaxul not set"));
    with
    | Aborting -> raise Aborting
    | _ -> err_dl := true
    );
  (* upload config error raising *)
    (try
      if res.ul_host = "" then raise (Grenouille_command ("get_config","ul_host not set"));
      if res.ul_path = "" then raise (Grenouille_command ("get_config","ul_path not set"));
      if res.ul_file = "" then raise (Grenouille_command ("get_config","ul_file not set"));
      if res.ul_size = "" then raise (Grenouille_command ("get_config","ul_size not set"));
      if res.ul_frequency = "" then raise (Grenouille_command ("get_config","ul_frequency not set"));
      if res.ul_max = "" then raise (Grenouille_command ("get_config","ul_max not set"));
      if res.ul_ratiomaxdl = "" then raise (Grenouille_command ("get_config","ul_rationmaxdl not set"));
      if res.ul_ratiomaxul = "" then raise (Grenouille_command ("get_config","ul_rationmaxul not set"));
    with
    | Aborting -> raise Aborting
    | _ -> err_ul := true
    );
  
    (* ping config error raising *)
    (try
      if res.ping_host = "" then raise (Grenouille_command ("get_config","ping_host not set"));
      if res.ping_quantity = "" then raise (Grenouille_command ("get_config","ping_quantity not set"));
      if res.ping_frequency = "" then raise (Grenouille_command ("get_config","ping_frequency not set"));
      if res.ping_min = "" then raise (Grenouille_command ("get_config","ping_min not set"));
      if res.ping_max = "" then raise (Grenouille_command ("get_config","ping_max not set"));
      if res.ping_string = "" then raise (Grenouille_command ("get_config","ping_string not set"));
    with
    | Aborting -> raise Aborting
    | _ -> err_ping := true
  );
  (* breakdown config error raising *)
    (try  
      if res.breakdown_host = "" then raise (Grenouille_command ("get_config","breakdown_host not set"));
      if res.breakdown_frequency = "" then raise (Grenouille_command ("get_config","breakdown_frequency not set"));
    with
    | Aborting -> raise Aborting
    | _ -> err_bk := true
  );
    if res.time_daylightsaving = "" then raise (Grenouille_command ("get_config","time_daylightsaving not set"));
    if res.time_zone = "" then raise (Grenouille_command ("get_config","time_zone not set"));
    
    if res.config_timetolive = "" then raise (Grenouille_command ("get_config","config_timetolive not set"));
   
  (* debug printing for now *)
    
    dprint Debug ("dl_host = " ^ res.dl_host);
    dprint Debug ("dl_path = " ^ res.dl_path);
    dprint Debug ("dl_file = " ^ res.dl_file);
    dprint Debug ("dl_frequency = " ^ res.dl_frequency);
    dprint Debug ("dl_ratiomaxdl = " ^ res.dl_ratiomaxdl);
    dprint Debug ("dl_ratiomaxul = " ^ res.dl_ratiomaxul);
    
    dprint Debug ("ul_host = " ^ res.ul_host);
    dprint Debug ("ul_path = " ^ res.ul_path);
    dprint Debug ("ul_file = " ^ res.ul_file);
    dprint Debug ("ul_size = " ^ res.ul_size);
    dprint Debug ("ul_frequency = " ^ res.ul_frequency);
    dprint Debug ("ul_max = " ^ res.ul_max);
    dprint Debug ("ul_ratiomaxdl = " ^ res.ul_ratiomaxdl);
    dprint Debug ("ul_ratiomaxul = " ^ res.ul_ratiomaxul);
    
    dprint Debug ("ping_host = " ^ res.ping_host);
    dprint Debug ("ping_quantity = " ^ res.ping_quantity);
    dprint Debug ("ping_frequency = " ^ res.ping_frequency);
    dprint Debug ("ping_min = " ^ res.ping_min);
    dprint Debug ("ping_max = " ^ res.ping_max);
    dprint Debug ("ping_string = " ^ res.ping_string);
    
    dprint Debug ("breakdown_host = " ^ res.breakdown_host);
    dprint Debug ("breakdown_frequency = " ^ res.breakdown_frequency);
    
    dprint Debug ("time_daylightsaving = " ^ res.time_daylightsaving);
    dprint Debug ("time_zone = " ^ res.time_zone);
    
    dprint Debug ("config_timetolive = " ^ res.config_timetolive);
    
    let dl_host = if usrconf.dlhost <> "" then usrconf.dlhost else res.dl_host in
    let dl_path = if usrconf.dlpath <> "" then usrconf.dlpath else res.dl_path in
    let dl_file = if usrconf.dlfile <> "" then usrconf.dlfile else res.dl_file in
    let ul_host = if usrconf.ulhost <> "" then usrconf.ulhost else res.ul_host in
    let ul_path = if usrconf.ulpath <> "" then usrconf.ulpath else res.ul_path in
    let ul_file = if usrconf.ulfile <> "" then usrconf.ulfile else res.ul_file in


    (* potential bug here ... when something fails for the first time
     * it is never tried again
     *)
    
    commands.download <-     
      (if (!err_dl) then fun _ -> dprint Normal "error in getting config for download prevents test" 
       else download dl_host dl_path dl_file (int_of_string res.dl_ratiomaxdl) (int_of_string res.dl_ratiomaxul) (int_of_string res.dl_frequency));

    commands.upload <- 
      (if (!err_ul) then fun _ -> dprint Normal "error in getting config for upload prevents test"
      else upload (float_of_int (int_of_string res.ul_max)) ul_host ul_path
      ul_file (int_of_string res.ul_size) (int_of_string res.ul_ratiomaxdl)
      (int_of_string res.ul_ratiomaxul) (int_of_string res.ul_frequency));
    
    commands.ping <- 
      (if (!err_ping) then fun _ -> dprint Normal "error in getting config for ping prevents test"
      else ping (int_of_string res.ping_frequency) (int_of_string res.dl_frequency) res.ping_host
      res.ping_quantity res.ping_string);

    commands.breakdown <- 
      (if (!err_bk) then fun _ -> dprint Normal "error in getting config for breakdown prevents test"
      else breakdown (generate_list_from_commas res.breakdown_host)
      (int_of_string res.breakdown_frequency));
    
    if (not (!err_dl)) then (commands.test_download <- generate_test_download
    (int_of_string res.dl_ratiomaxdl) (int_of_string res.dl_ratiomaxul));

    if (not (!err_ul)) then (commands.test_upload <- generate_test_upload
    (int_of_string res.ul_ratiomaxdl) (int_of_string res.ul_ratiomaxul));

    if !connection_failures > 0 then begin
      let delay = (!min_time / 10) * ((Random.int 10) + 1) in
      let date = Future.set_time delay in
      Timed_events.delay_connections true date delay
    end;
    connection_failures := 0;
    Timed_events.add_event_unique Get_config (Future.set_time (int_of_string res.config_timetolive))

    end (* if not using alt *)
  
  with 
  | Aborting -> raise Aborting
  | e ->
      dprint Normal ("Problem getting config:\n" ^ Printexc.to_string e);
      incr connection_failures;
      let delay = compute_delay () in
      let date = Future.set_time delay in
      Timed_events.add_event_unique Get_config date;
      Timed_events.delay_connections false date delay

let post_ping id sent received total_time elapsed use_alt =
  let cmd =
    "post_ping&id=" ^ id ^
    "&sent=" ^ (string_of_int sent) ^
    "&received=" ^ (string_of_int received) ^
    "&total_time=" ^ (string_of_float total_time) ^
    "&elapsed_seconds=" ^ (string_of_int (int_of_float elapsed))
  in
  process_command cmd use_alt
      
let post_ul id bandwidth elapsed use_alt =
  let cmd =
    "post_ul&id=" ^ id ^
    "&bandwidth=" ^ (string_of_float bandwidth) ^
    "&elapsed_seconds=" ^ (string_of_int (int_of_float elapsed))
  in
  process_command cmd use_alt

let post_dl id bandwidth elapsed use_alt =
  let cmd =
    "post_dl&id=" ^ id ^
    "&bandwidth=" ^ (string_of_float bandwidth) ^
    "&elapsed_seconds=" ^ (string_of_int (int_of_float elapsed))
  in
  process_command cmd use_alt

let post_breakdown id elapsed use_alt =
  let cmd =
    "post_breakdown&id=" ^ id ^
    "&elapsed_seconds=" ^ (string_of_int (int_of_float elapsed))
  in
  process_command cmd use_alt

exception Failed_create_command

let create_commands () =
  if usrconf.alt_grenouille_server <> "" then begin
    dprint Debug ("Sending a duplicate of results to: "^ usrconf.alt_grenouille_server);
    use_alt_server := true
  end else begin
    dprint Debug "Not sending duplicate of results";
    use_alt_server := false
  end;
  try
    if !use_alt_server then begin
      let pass = Crypt.crypt usrconf.password in 
      say_hello pass true;
      get_config !alt_id true;
    end;
    if (!id = "") then
      (let pass = Crypt.crypt usrconf.password in 
      say_hello pass false);
    myign get_last_version !id;
    commands.get_config <- get_config !id;
    commands.post_ping <- post_ping !id;
    commands.post_ul <- post_ul !id;
    commands.post_dl <- post_dl !id;
    commands.post_breakdown <- post_breakdown !id;
    dprint Normal "getting config";
    commands.get_config false;
    (if usrconf.testdl then
      Timed_events.add_event_unique DoDownload (Future.set_time 0)
    else
      dprint Debug "config file says not to test download");
    (if usrconf.testping then
      Timed_events.add_event_unique DoPing (Future.set_time 1)
    else
      dprint Debug "config file says not to test ping");
    (if usrconf.testul then
      Timed_events.add_event_unique DoUpload (Future.set_time 2)
    else
      dprint Debug "config file says not to test upload");
    (if usrconf.testbk then
      Timed_events.add_event_unique DoBreakdown (Future.set_time 3)
    else
      dprint Debug "config file says not to test breakdown");  
    Timed_events.add_event_unique PostResults (Future.set_time 10)
  with
  | Aborting -> raise Aborting
  | e ->
      dprint Normal ("Problem creating commands:\n" ^ Printexc.to_string e);
      raise Failed_create_command
