(***************************************************************)
(*                                                             *)
(*                 camlgrenouille for Linux                    *)
(*                                                             *)
(*             Copyright 2001-2003 Alan Schmitt                *)
(*              alan.schmitt@polytechnique.org                 *)
(*                                                             *)
(* This file is distributed under the terms of the GNU General *)
(* Public License.                                             *)
(*                                                             *)
(***************************************************************)

(*
open Concur
open ThreadUnix
*)
open Debug
open Abort

exception Cannot_build_data_connection
  
type control_connection = {
    cin : in_channel; 
    cout : out_channel;
  }

(* to authentify and work with the connection *)
type connection_automata_state =
  | UserState
  | PassState
  | InitState

type connect_args =
  Port of int
| User of string
| Password of string

type get_args =
  Passive

type put_args =
  PassivePut

module type SysCalls = sig
    val open_connection : Unix.sockaddr -> in_channel * out_channel
    val connect :  Unix.file_descr -> Unix.sockaddr -> unit
    val listen : Unix.file_descr -> int -> unit
    val bind : Unix.file_descr -> Unix.sockaddr -> unit
    val accept : Unix.file_descr -> Unix.file_descr * Unix.sockaddr
    val getsockname :  Unix.file_descr -> Unix.sockaddr
    val setsockopt : Unix.file_descr -> Unix.socket_bool_option -> bool -> unit
    val socket :  Unix.socket_domain -> Unix.socket_type -> int -> 
      Unix.file_descr
    val close : Unix.file_descr -> unit
  end
      
module Make = functor (SysCalls: SysCalls) -> struct

let timeout = ref 20.0

let set_timeout time =
  timeout := time

(* opening the link *)

let create_link ip port =
  Timeout.timeout !timeout SysCalls.open_connection
    (Unix.ADDR_INET (ip, port))

(* get the answer from the control_in link *)

let find_code s = try String.sub s 0 3  with Invalid_argument _ -> ""

let is_last s code =
  try
    ((String.sub s 0 4) = (code ^" ")) & 
    (try (int_of_string (String.sub s 0 3)) < 600 with Failure s -> false)
  with 
  | Invalid_argument _ -> false


let receive_answer control_in =
  let rec get_rest code res =
    let line = Timeout.timeout !timeout input_line control_in in
    dprint Debug line;
    if is_last line code then code, (line ^ "\n" ^ res)
    else get_rest code (line ^ "\n" ^ res)
  in
  try
    let line = Timeout.timeout !timeout input_line control_in in
    dprint Debug line;
    let code = find_code line in
    if String.sub line 3 1 = "-" then
      get_rest code line
    else code,line
  with Timeout.Timeout -> "Timeout","Timeout"

let do_connection cin cout user password=
  let send_user () =
    dprint Debug ("USER "^user^"\n");
    output_string cout ("USER "^user^"\r\n");
    flush cout;
    receive_answer cin
  in
  let send_pass () =
    dprint Debug ("PASS "^password^"\n");
    output_string cout ("PASS "^password^"\r\n");
    flush cout;
    receive_answer cin
  in
  let rec connection_automata = function
    | UserState -> (
	match send_user () with
	| "331",_ -> connection_automata PassState
	| "332",_ -> connection_automata PassState
	| "230",_ -> connection_automata InitState
	| s,_ -> s )
    | PassState -> (
	match send_pass () with
	| "230",_ -> connection_automata InitState
	| "202",_ -> connection_automata InitState
	| s,_ -> s )
    | InitState -> (
	output_string cout "TYPE I\r\n";
	flush cout;
	let _ = receive_answer cin in
	"connected" )
  in
  connection_automata UserState

let get_ip host =
  try
    (Unix.gethostbyname host).Unix.h_addr_list.(0)
  with
  | Not_found -> failwith "could not resolve ip address"

let login = try Sys.getenv "USER" with _ -> "nobody"
let host = 
  let h =  Unix.gethostname () in
  if String.contains h '.' then h else
    (* Unix2 is not included anymore, dirty hack
    h ^ "." ^ (Unix2.getdomainname ()) *)
    "nowhere.com"

let mail_address =
  login ^ "@" ^ host
  
let connect host args =
  let port = ref 21 in
  let password = ref mail_address in
  let user = ref "anonymous" in
  (*
    Alan: removed next line, doesn't seem to do anything
  let timeout = ref !timeout in
  *)
  List.iter (function
      Port p -> port := p
    | User s -> user := s
    | Password s -> password := s
  ) args;
  let password = !password in
  let user = !user in
  let port = !port in
  let ip_addr = get_ip host in
  dprint Debug ("Try to connect to : "^(Unix.string_of_inet_addr ip_addr));
  let cin,cout = create_link ip_addr port in
  dprint Debug "link created";
  let code,res =
    try
      receive_answer cin
    with
    | e -> 
	myign close_out cout; 
	raise e
  in
  match code with
  | "220" -> (
      match 
	try do_connection cin cout user password
	with
	| e -> 
	    myign close_out cout; 
	    raise e
      with
      |	"connected" -> 
	  dprint Debug "connected";
	  { cin = cin; cout = cout}
      |	s -> 
	  myign close_out cout; 
	  failwith ("Connection failed: "^s) )
  | s -> 
      myign close_out cout; 
      failwith ("Connection failed: "^s)
	
let quit conn =
  (try
    output_string conn.cout "QUIT\r\n";
    dprint Debug "QUIT";
    flush conn.cout;
    let _ = receive_answer conn.cin in
    ()
  with
  | e -> 
      myign close_out conn.cout; 
      raise e);
  myign close_out conn.cout
    
let size conn file=
  output_string conn.cout ("SIZE "^file^"\r\n");
  flush conn.cout;
  let code,res = receive_answer conn.cin in
  match code with 
  | "213" -> int_of_string (String.sub res 4 ((String.length res) - 5))
  | _ -> failwith ("Could not get size "^res)

(* function to deal with data connections *)

let data_port = ref 10000

let get_data_port () =
  if !data_port > 20000 then data_port := 10000;
  let n = !data_port in
  incr data_port;
  n

exception Addr_found of Unix.inet_addr
exception No_inet_addr

let get_my_inet_addr conn =
  let fd = Unix.descr_of_out_channel conn.cout in
  let addr = SysCalls.getsockname fd in
  match addr with
  | Unix.ADDR_UNIX _ -> failwith "Unable to get local ip address: unix socket"
  | Unix.ADDR_INET (inet_addr, port) -> 
      dprint Debug ("my ip address is: "^(Unix.string_of_inet_addr inet_addr));
      inet_addr

(* beware, this function puts a comma at the end of the string *)
let stupid_dot_to_comma_conversion_with_ending_comma s =
  let rec f s =
    try
      let n = String.index s '.' in
      (String.sub s 0 n) :: (f (String.sub s (n+1) ((String.length s) - n - 1)))
    with
      Not_found -> [s]
  in
  (* beware, comma at the end of the string *)
  List.fold_right (fun s -> fun sl -> s ^ "," ^ sl) (f s) ""

exception Got_port of int

let try_several_bind my_addr sock n =
  try
    for i = 1 to n do
      try
	let dp = get_data_port () in
        dprint Debug ("trying port "^(string_of_int dp));
	SysCalls.bind sock (Unix.ADDR_INET (my_addr, dp));
	raise (Got_port dp)
      with
      | (Got_port d) as e -> raise e
      | Unix.Unix_error (Unix.EINVAL,_,_) -> 
	  dprint Debug "port already bound, trying again"
      | Unix.Unix_error (Unix.EADDRINUSE,_,_) -> 
	  dprint Debug "address already bound, trying another port"
      | (Unix.Unix_error (e, _, _)) as e' -> 
          dprint Debug ("bind Unix error: "^(Unix.error_message e)); raise e'
      | e -> dprint Debug ("bind error: "^(Printexc.to_string e)); raise e
    done;
    failwith ("Could not get port to bind to address after "^(string_of_int n)^" tries")
  with
  | Got_port dp -> dp
  | e -> raise e

let new_data_connection connection = 
  let sock = SysCalls.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
  try
    SysCalls.setsockopt sock Unix.SO_REUSEADDR true;
    let my_addr = get_my_inet_addr connection
    in
    let dp = try_several_bind my_addr sock 10 in
    SysCalls.listen sock 5;
    let set_data_port addr dp conn =
      let s = stupid_dot_to_comma_conversion_with_ending_comma (Unix.string_of_inet_addr addr) in
      let dp2 = (string_of_int (dp mod 256)) in
      let dp1 = (string_of_int (dp / 256)) in
      let cmd = "PORT " ^ s ^ dp1 ^ "," ^ dp2 ^ "\r\n" in
      dprint Debug cmd;
      output_string conn.cout cmd;
      flush conn.cout;
      let code,res = receive_answer conn.cin in
      code
    in
    match set_data_port my_addr dp connection with
    | "200" -> sock
    | _ -> failwith "problem getting a data port"
  with
    e -> myign SysCalls.close sock; raise e

exception Retrieve_failed of string

let get_data = Bufrw.buf_read 

let put_data = Bufrw.buf_write 

let exec_data_command cmd connection how func_to_apply get_flag pass_flag=
  let data = 
    if pass_flag then
      SysCalls.socket Unix.PF_INET Unix.SOCK_STREAM 0
    else
      new_data_connection connection
  in
  let t_start,t0,t1,t2 =  
    try
      if pass_flag then (
	output_string connection.cout ("PASV\r\n");
	flush connection.cout;
	let code,res = receive_answer connection.cin in
	(match code with
	| "227" -> ()
	| _ -> failwith "could not enter passive mode");
	let ind = String.index res '(' in
	let ind2 = String.index res ')' in
	let addr_str = String.sub res (ind+1) (ind2 - ind - 1) in
	let get_next_dot s =
	  let ind = String.index s ',' in
	  let res = String.sub s 0 ind in
	  let rest = String.sub s (ind+1) ((String.length s) - ind - 1) in
	  res, rest
	in
	let addr1, r1 = get_next_dot addr_str in
	let addr2, r2 = get_next_dot r1 in
	let addr3, r3 = get_next_dot r2 in
	let addr4, r4 = get_next_dot r3 in
	let port1, port2 = get_next_dot r4 in
	let addr = addr1 ^ "." ^ addr2 ^ "." ^ addr3 ^ "." ^addr4 in
	let port = (int_of_string port1) * 256 + (int_of_string port2) in
	dprint Debug ("passive mode on address "^ addr ^ ":" ^ (string_of_int port));
	SysCalls.connect data (Unix.ADDR_INET ((Unix.inet_addr_of_string addr), port));
       );
  (* here starts exec_data_command code *)
      output_string connection.cout (cmd ^ "\r\n");
      dprint Debug cmd;
      flush connection.cout;
      let tt_start = Unix.gettimeofday () in
      let code,res = receive_answer connection.cin in
      match (String.sub code 0 1) with
      | "4" | "5" -> (
	  match code with
	  | "425" -> raise Cannot_build_data_connection
	  | _ -> raise (Retrieve_failed res)
	 )
      | _ ->
	  let tt0 = Unix.gettimeofday () in
	  let data_tmp =
	    if pass_flag then
	      data
	    else
	      fst (Timeout.timeout !timeout SysCalls.accept data)
	  in
	  try
	    let tt1 = Unix.gettimeofday () in 
	    how data_tmp func_to_apply;
	    let tt2 = Unix.gettimeofday () in
	    myign SysCalls.close data_tmp;
	    tt_start,tt0,tt1,tt2
	  with
	  | e -> myign SysCalls.close data_tmp; raise e
    with
    | e -> myign SysCalls.close data; raise e
  in
  if (not pass_flag) then
    myign SysCalls.close data;
  let t3 = Unix.gettimeofday () in
  let code,res = receive_answer connection.cin in 
  match code with
  | "226" -> 
      let t4 = Unix.gettimeofday () in
      dprint Debug ("send_req: "^(string_of_float (t0 -. t_start))^" accept: "^(string_of_float (t1 -. t0))^" data transfer: "^(string_of_float (t2 -. t1))^" close: "^(string_of_float (t3 -. t2))^" end: "^(string_of_float (t4 -. t3)));
      if get_flag 
      then t2 -. t_start
      else t4 -. t1
  | _ -> failwith ("problem while transfering "^res)

exception Get_ok of float

let repeat_exec_data_command_n_times cmd connection how func_to_apply get_flag pass_flag n =
  try
    for i = 1 to 10 do
      try
	let t = exec_data_command cmd connection how func_to_apply get_flag pass_flag in
	raise (Get_ok t)
      with
      | Cannot_build_data_connection -> ()
      | e -> raise e
    done;
    failwith ("could not open data connection after "^(string_of_int n)^" tries")
  with
  | Get_ok time -> time
      
let get connection file read_buffer args =
  let pass_flag = ref false in
  List.iter (function Passive -> pass_flag := true) args;
  let pass_flag = !pass_flag in
  let cmd = "RETR" ^ (if file <> "" then " " else "") ^ file in
  repeat_exec_data_command_n_times cmd connection 
  (get_data (Bufrw.Time !timeout)) read_buffer true pass_flag 10
  
let put connection my_fill_buffer file args =
  let pass_flag = ref false in
  List.iter (function PassivePut -> pass_flag := true) args;
  let pass_flag = !pass_flag in
  let cmd = "STOR " ^ file in
  repeat_exec_data_command_n_times cmd connection (put_data !timeout) my_fill_buffer false pass_flag 10
    
let cd conn path =
  output_string conn.cout ("CWD"^ (if path <> "" then " " ^ path else "") ^ "\r\n");
  flush conn.cout;
  let code,res = receive_answer conn.cin in
  match code with
  | "250" -> ()
  | _ -> failwith ("could not cd: "^res)

let rm conn file =
  output_string conn.cout ("DELE"^ (if file <> "" then " " ^ file else "") ^ "\r\n");
  flush conn.cout;
  let code,res = receive_answer conn.cin in
  match code with
  | "250" -> ()
  | _ -> failwith ("could not rm: "^res)

end

module FtpUnix = Make(
    struct 
      include Unix (* standard Unix calls *)
      (*include ThreadUnix (* Unix calls with threads *)*)
    end)
  
let put = FtpUnix.put
let get = FtpUnix.get
let size = FtpUnix.size
let rm = FtpUnix.rm
let cd = FtpUnix.cd
let quit = FtpUnix.quit
let connect = FtpUnix.connect
let set_timeout = FtpUnix.set_timeout
