(***************************************************************)
(*                                                             *)
(*                      camlgrenouille                         *)
(*                                                             *)
(*             Copyright 2001-2005 Alan Schmitt                *)
(*              alan.schmitt@polytechnique.org                 *)
(*                                                             *)
(* This file is distributed under the terms of the GNU General *)
(* Public License.                                             *)
(*                                                             *)
(***************************************************************)

open Commands
open Debug
open Config    
open Sysconf
open Abort

let config_file = ref "user.config"
let file_in_name = ref ""
let test_ping = ref ""
let do_print_time = ref false

(* For testing purposes *)
let dummy_mode = ref false

(* This is set to true by the user when there is something to read on file_in_name *)
let incoming_command = ref false

let print_time s =
  let time = Unix.localtime (Unix.time ()) in
  let _, time' = Unix.mktime {time with Unix.tm_sec = time.Unix.tm_sec + s} in
  Printf.sprintf "[%04d-%02d-%02d %02d:%02d:%02d]"
    (time'.Unix.tm_year + 1900)
    (time'.Unix.tm_mon + 1)
    time'.Unix.tm_mday
    time'.Unix.tm_hour
    time'.Unix.tm_min
    time'.Unix.tm_sec

let main_loop () = 
  let lex = function
    | "dummy" -> 
        dprint Debug "commande dummy comprise";
        Some (Event_list.DoDummy)
    | s ->
        dprint Debug ("commande NON COMPRISE: "^s);
        None
  in
  let process_incoming_commands pipe =
    incoming_command := false;
    try
      while true do
        let s = input_line pipe in
          match (lex s) with
            | None -> ()
            | Some e -> Timed_events.add_event e (Future.set_time 0)
      done
    with
      | End_of_file -> ()
  in
	let event_wait () =
		let t = Timed_events.time_to_next_event () in
		let wh = if !do_print_time then (" (at "^(print_time t)^")") else "" in
		dprint Normal "";
		dprint Normal ("Prochain test dans "^(string_of_int t)^" secondes" ^ wh ^": "^(Timed_events.string_of_next_event ()));
		if t <> 0 then Unix.sleep t
	in
	let rec loop_no_pipe () =
		event_wait ();
		let ev = Timed_events.get_first_event () in
		Event.do_event ev;
		loop_no_pipe ()
	in
	let rec loop_pipe pipe =
		event_wait ();
		if !incoming_command then begin
			dprint Debug "Commandes en attente ...";
			process_incoming_commands pipe;
			loop_pipe pipe
		end else
			let ev = Timed_events.get_first_event () in
			Event.do_event ev;
			loop_pipe pipe
	in
	if !file_in_name = "" then loop_no_pipe ()
	else 
		let pipe = open_in_gen [Open_nonblock] 0o640 !file_in_name in
		dprint Debug "pipe in ouverte";
		loop_pipe pipe

exception OK

let int_handler = Sys.Signal_handle
    (fun _ ->
      Sys.set_signal Sys.sigint Sys.Signal_default;
    raise Aborting
    )

let _ =
  Random.self_init();
  Arg.parse
    [
      "-f", Arg.String (fun s -> config_file := s), ("<fichier>: utiliser le fichier de configuration <fichier> défaut: "^(!config_file)) ;
      "-q", Arg.Unit (fun () -> set_debug_level Quiet ), "mode silencieux" ;
      "-d", Arg.Unit (fun () -> set_debug_level Debug), "mode très verbeux" ;
      "-t", Arg.Unit (fun () -> set_debug_prefix (fun () -> print_time 0)), "imprime l'heure courante avant tout message";
      "-at", Arg.Unit (fun () ->  do_print_time := true), "imprime l'heure du test suivant";
      "--test_ping", Arg.String (fun s -> test_ping :=s), "<hôte>: teste le ping avec <hôte>";
      "--dummy_mode", Arg.Unit (fun () -> dummy_mode := true), "lance camlgrenouille avec seulement un événement dummy (test)"
    ]
    (fun s -> ())
    "";
  dprint Normal ("Bonjour. Je suis camlgrenouille version "^sysconf.version_maj ^ "." ^ sysconf.version_min ^
	  (match sysconf.beta with None -> ""
	  | Some s -> " beta "^s));
  dprint Normal "camlgrenouille est sous le copyright 2001-2006 de Alan Schmitt et Romain Cledat";
  dprint Normal "et est publié sous la licence GNU public license";
  dprint Normal "(voir le fichier LICENSE ou www.gnu.org pour les détails)";
  dprint Normal "Le portage vers OS X a été fait par Romain Cledat";
  if not (parse_config !config_file) then exit 1;
	(* On se prépare à recevoir des données par socket UNIX de l'extérieur *)
	if !file_in_name <> "" then begin
		let usr1_handler = Sys.Signal_handle
			(fun _ ->
				incoming_command := true
				)
		in
		Sys.set_signal Sys.sigusr1 usr1_handler;
		if Sys.file_exists !file_in_name then Sys.remove !file_in_name;
		Unix.mkfifo !file_in_name 0o640
	end;
  Sys.set_signal Sys.sigint int_handler;
  if (!test_ping) <> "" then
    let trans, reic, time = Utils.ping (!test_ping) "10" "0" in
    dprint Normal ("ping: envoyés: "^(string_of_int trans)^"; reçus: "^(string_of_int reic)^"; temps moyen: "^(string_of_float (time/. (float_of_int reic)))^" ms");
    exit 0
  else if !dummy_mode then begin
    Timed_events.add_event Event_list.DoDummy (Future.set_time 1000);
    main_loop ()
  end else
  try
    while true do
      try
	create_commands ();
	dprint Debug "create_command OK";
	raise OK
      with
      |	Failed_create_command -> Unix.sleep 300;
    done;
  with
  | OK -> 
      try 
	main_loop ()
      with
      |	Aborting -> 
          if not usrconf.post_results then
	    dprint Normal "Signal de fin reçu, configuration ne veut pas d'envoi de résultats."
          else begin
            dprint Normal "Signal de fin reçu, envoi des résultats.";
            while (!results_to_send <> [] && usrconf.post_results) do
              do_post_results ();
              if !results_to_send <> [] then Unix.sleep 5
            done;
          end;
          (* Cleaning up *)
          if !file_in_name <> "" then Sys.remove !file_in_name;
	  dprint Normal "Au revoir.";
	  exit 0